from typing import List, Optional, cast
import strawberry
from django.db.models import fields, Model
from . import utils

field_type_map = {
    fields.AutoField: strawberry.ID,
    fields.IntegerField: int,
    fields.CharField: str,
    fields.TextField: str,
    fields.BooleanField: bool,
}

def get_field_type(field):
    db_field_type = type(field)
    return field_type_map.get(db_field_type)


model_type_map = {
}

def set_model_type(model, field_type):
    model_type_map[model] = field_type

def get_model_type(model):
    model_type = model_type_map.get(model)
    if model_type is None:
        model_type = LazyModelType(model)
    return model_type

class LazyModelType(strawberry.LazyType):
    def __init__(self, model):
        super().__init__('', '', '')
        self.model = model
    def resolve_type(self):
        if self.model not in model_type_map:
            raise Exception(f'GraphQL type not defined for "{self.model._meta.object_name}" Django model.')
        return model_type_map[self.model]


def get_field(field, is_input, is_update):
    field_type = field_type_map.get(type(field))

    if is_input and field_type == strawberry.ID:
        return #TODO: is this correct?

    if field_type is None:
        print('Unknown field type', type(field))
        return

    if is_input:
        if field.blank or is_update:
            field_type = Optional[field_type]

    return field.name, field_type, strawberry.arguments.UNSET


def get_relation_field(field):
    if hasattr(field, 'get_accessor_name'):
        field_name = field.get_accessor_name()
    else:
        field_name = field.name
    field_type = get_model_type(field.related_model)
    @strawberry.field
    def resolver(info, root,
            filter: Optional[List[str]] = None,
            exclude: Optional[List[str]] = None) -> List[field_type]:
        qs = getattr(root, field_name).all()
        return utils.filter_qs(qs, filter, exclude)

    return field_name, None, resolver


def get_relation_foreignkey_field(field):
    field_name = field.name
    field_type = get_model_type(field.related_model)

    @strawberry.field
    def resolver(info, root) -> Optional[field_type]:
        obj = getattr(root, field_name)
        return obj

    return field_name, None, resolver


def generate_model_type(resolver, is_input=False, is_update=False):
    model = resolver.model
    annotations = {}
    attributes = { '__annotations__': annotations }
    resolver_fields = getattr(resolver, 'fields', None)
    resolver_exclude = getattr(resolver, 'exclude', None)

    # add fields
    for field in model._meta.get_fields():
        if resolver_fields and field.name not in resolver_fields:
            continue # skip
        if resolver_exclude and field.name in resolver_exclude:
            continue # skip
        if field.is_relation:
            if is_input:
                continue
            if isinstance(field, fields.related.ForeignKey):
                field_params = get_relation_foreignkey_field(field)
            else:
                field_params = get_relation_field(field)
        else:
            field_params = get_field(field, is_input, is_update)
        if not field_params:
            continue
        field_name, field_type, field_value = field_params
        attributes[field_name] = field_value
        if field_type:
            annotations[field_name] = field_type
    if not is_input:
        for field_name in dir(resolver):
            field = getattr(resolver, field_name)
            if hasattr(field, '_field_definition'):
                attributes[field_name] = field

    # generate type
    type_name = model._meta.object_name
    if is_update:
        type_name = f'Update{type_name}'
    elif is_input:
        type_name = f'Create{type_name}'
    model_type = type(type_name, (), attributes)
    model_type = strawberry.type(model_type, is_input=is_input)
    if not is_input:
        set_model_type(model, model_type)
    return model_type



from typing import List, Optional, cast
import strawberry
from . import utils
from .types import generate_model_type
from django.core import exceptions


class ModelResolverMixin:
    def __init__(self, info, root):
        self.info = info
        self.root = root
        self.request = None
        if info.context:
            self.request = info.context.get('request')

    def get_model(self):
        return self.model

    def get_queryset(self):
        model = self.get_model()
        return model.objects.all()

    def get_queryset_filtered(self, filter):
        qs = self.get_queryset()
        if not filter:
            return qs
        filter, exclude = utils.split_filters(filter)
        if filter:
            qs = qs.filter(**filter)
        if exclude:
            qs = qs.exclude(**exclude)
        return qs

    def list(self, **filters):
        self.check_permissions('view')
        qs = self.get_queryset_filtered(**filters)
        return qs

    def get(self, id):
        self.check_permissions('view')
        qs = self.get_queryset()
        return qs.get(id=id)

    def create(self, data):
        self.check_permissions('add')
        model = self.get_model()
        instance = model(**data)
        instance.save()
        return instance

    def update(self, data, **filters):
        self.check_permissions('change')
        qs = self.get_queryset_filtered(**filters)
        qs.update(**data)
        return qs

    def delete(self, **filters):
        self.check_permissions('delete')
        qs = self.get_queryset_filtered(**filters)
        items = list(qs)
        qs.delete()
        return items

    def check_permissions(self, permission_codenames):
        _super = super()
        if hasattr(_super, 'check_permissions'):
            _super.check_permissions(permission_codenames)


class ModelPermissionMixin:
    def check_permissions(self, permission_codenames):
        if not self.request:
            return
        if isinstance(permission_codenames, str):
            permission_codenames = [permission_codenames]
        app_label = self.model._meta.app_label
        model_name = self.model._meta.model_name
        perms = [f'{app_label}.{perm}_{model_name}' for perm in permission_codenames]
        if not self.request.user.has_perms(perms):
            raise exceptions.PermissionDenied('Permission denied')


class ModelFieldMixin:
    @classmethod
    def get_field(cls):
        @strawberry.field
        def get_field(info, root, id: strawberry.ID) -> cls.output_type:
            instance = cls(info, root)
            return instance.get(id)
        return get_field

    @classmethod
    def list_field(cls):
        @strawberry.field
        def list_field(info, root,
                filter: Optional[List[str]] = None) -> List[cls.output_type]:
            instance = cls(info, root)
            return instance.list(filter=filter)
        return list_field

    @classmethod
    def query_fields(cls):
        return cls.get_field(), cls.list_field()


class ModelMutationMixin:
    @classmethod
    def create_mutation(cls):
        @strawberry.mutation
        def create_mutation(info, root, data: cls.create_input_type) -> cls.output_type:
            instance = cls(info, root)
            return instance.create(utils.get_data(cls.model, data))
        return create_mutation

    @classmethod
    def update_mutation(cls):
        @strawberry.mutation
        def update_mutation(info, root, data: cls.update_input_type,
                filter: Optional[List[str]] = None) -> List[cls.output_type]:
            instance = cls(info, root)
            return instance.update(utils.get_data(cls.model, data), filter=filter)
        return update_mutation

    @classmethod
    def delete_mutation(cls):
        @strawberry.mutation
        def delete_mutation(info, root,
                filter: Optional[List[str]] = None) -> List[cls.output_type]:
            instance = cls(info, root)
            return instance.delete(filter=filter)
        return delete_mutation

    @classmethod
    def mutations(cls):
        return cls.create_mutation(), cls.update_mutation(), cls.delete_mutation()


class ModelQueryMixin:
    @classmethod
    def query(cls):
        verbose_name = cls.model._meta.verbose_name
        class Query: pass
        setattr(Query, f'{verbose_name}', cls.get_field())
        setattr(Query, f'{verbose_name}s', cls.list_field())
        return strawberry.type(Query)

    @classmethod
    def mutation(cls):
        verbose_name = cls.model._meta.verbose_name
        class Mutation: pass
        setattr(Mutation, f'create_{verbose_name}', cls.create_mutation())
        setattr(Mutation, f'update_{verbose_name}s', cls.update_mutation())
        setattr(Mutation, f'delete_{verbose_name}s', cls.delete_mutation())
        return strawberry.type(Mutation)


class ModelTypeBase(type):
    def __new__(cls, name, bases, attrs, **kwargs):
        new_cls = super().__new__(cls, name, bases, attrs, **kwargs)
        if hasattr(new_cls, 'model'):
            new_cls.output_type = generate_model_type(new_cls)
            new_cls.create_input_type = generate_model_type(new_cls, is_input=True)
            new_cls.update_input_type = generate_model_type(new_cls, is_input=True, is_update=True)
        return new_cls


class ModelResolver(
        ModelResolverMixin,
        ModelPermissionMixin,
        ModelFieldMixin,
        ModelMutationMixin,
        ModelQueryMixin,
        metaclass=ModelTypeBase):
    pass
